@@ -83,8 +83,9 @@ class Backend():
                 return i
         raise RuntimeError('No compiler for language ' + lang)
 
-    def get_compiler_for_source(self, src):
-        for i in self.build.compilers:
+    def get_compiler_for_source(self, src, is_cross):
+        comp = self.build.cross_compilers if is_cross else self.build.compilers
+        for i in comp:
             if i.can_compile(src):
                 return i
         if isinstance(src, mesonlib.File):
@@ -92,6 +93,7 @@ class Backend():
         raise RuntimeError('No specified compiler can handle file ' + src)
 
     def get_target_filename(self, target):
+        assert(isinstance(target, (build.BuildTarget, build.CustomTarget)))
         targetdir = self.get_target_dir(target)
         fname = target.get_filename()
         if isinstance(fname, list):
@@ -101,13 +103,20 @@ class Backend():
         filename = os.path.join(targetdir, fname)
         return filename
 
+    def get_target_filename_abs(self, target):
+        return os.path.join(self.environment.get_build_dir(), self.get_target_filename(target))
+
     def get_target_filename_for_linking(self, target):
         # On some platforms (msvc for instance), the file that is used for
         # dynamic linking is not the same as the dynamic library itself. This
         # file is called an import library, and we want to link against that.
-        # On platforms where this distinction is not important, the import
-        # library is the same as the dynamic library itself.
-        return os.path.join(self.get_target_dir(target), target.get_import_filename())
+        # On all other platforms, we link to the library directly.
+        if isinstance(target, build.SharedLibrary):
+            link_lib = target.get_import_filename() or target.get_filename()
+            return os.path.join(self.get_target_dir(target), link_lib)
+        elif isinstance(target, build.StaticLibrary):
+            return os.path.join(self.get_target_dir(target), target.get_filename())
+        raise AssertionError('BUG: Tried to link to something that\'s not a library')
 
     def get_target_dir(self, target):
         if self.environment.coredata.get_builtin_option('layout') == 'mirror':
@@ -129,7 +138,7 @@ class Backend():
         abs_files = []
         result = []
         for src in unity_src:
-            comp = self.get_compiler_for_source(src)
+            comp = self.get_compiler_for_source(src, target.is_cross)
             language = comp.get_language()
             suffix = '.' + comp.get_default_suffix()
             if language not in langlist:
@@ -226,7 +235,10 @@ class Backend():
 
     def determine_linker(self, target, src):
         if isinstance(target, build.StaticLibrary):
-            return self.build.static_linker
+            if self.build.static_cross_linker is not None:
+                return self.build.static_cross_linker
+            else:
+                return self.build.static_linker
         if len(self.build.compilers) == 1:
             return self.build.compilers[0]
         # Currently a bit naive. C++ must
@@ -284,6 +296,33 @@ class Backend():
             args = includeargs + args
         return args
 
+    @staticmethod
+    def escape_extra_args(compiler, args):
+        # No extra escaping/quoting needed when not running on Windows
+        if not mesonlib.is_windows():
+            return args
+        extra_args = []
+        # Compiler-specific escaping is needed for -D args but not for any others
+        if compiler.get_id() == 'msvc':
+            # MSVC needs escaping when a -D argument ends in \ or \"
+            for arg in args:
+                if arg.startswith('-D') or arg.startswith('/D'):
+                    # Without extra escaping for these two, the next character
+                    # gets eaten
+                    if arg.endswith('\\'):
+                        arg += '\\'
+                    elif arg.endswith('\\"'):
+                        arg = arg[:-2] + '\\\\"'
+                extra_args.append(arg)
+        else:
+            # MinGW GCC needs all backslashes in defines to be doubly-escaped
+            # FIXME: Not sure about Cygwin or Clang
+            for arg in args:
+                if arg.startswith('-D') or arg.startswith('/D'):
+                    arg = arg.replace('\\', '\\\\')
+                extra_args.append(arg)
+        return extra_args
+
     def generate_basic_compiler_args(self, target, compiler):
         commands = []
         commands += self.get_cross_stdlib_args(target, compiler)
@@ -292,7 +331,7 @@ class Backend():
         commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
         commands += self.build.get_global_args(compiler)
         commands += self.environment.coredata.external_args[compiler.get_language()]
-        commands += target.get_extra_args(compiler.get_language())
+        commands += self.escape_extra_args(compiler, target.get_extra_args(compiler.get_language()))
         commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
         if self.environment.coredata.get_builtin_option('werror'):
             commands += compiler.get_werror_args()
@@ -473,24 +512,34 @@ class Backend():
             if hasattr(i, 'held_object'):
                 i = i.held_object
             if isinstance(i, str):
-                fname = os.path.join(self.build_to_src, target.subdir, i)
+                fname = [os.path.join(self.build_to_src, target.subdir, i)]
             elif isinstance(i, build.BuildTarget):
-                fname = self.get_target_filename(i)
+                fname = [self.get_target_filename(i)]
+            elif isinstance(i, build.GeneratedList):
+                fname = [os.path.join(self.get_target_private_dir(target), p) for p in i.get_outfilelist()]
             else:
-                fname = i.rel_to_builddir(self.build_to_src)
+                fname = [i.rel_to_builddir(self.build_to_src)]
             if absolute_paths:
-                fname = os.path.join(self.environment.get_build_dir(), fname)
-            srcs.append(fname)
+                fname =[os.path.join(self.environment.get_build_dir(), f) for f in fname]
+            srcs += fname
         cmd = []
         for i in target.command:
             if isinstance(i, build.Executable):
                 cmd += self.exe_object_to_cmd_array(i)
                 continue
-            if isinstance(i, build.CustomTarget):
+            elif isinstance(i, build.CustomTarget):
                 # GIR scanner will attempt to execute this binary but
                 # it assumes that it is in path, so always give it a full path.
                 tmp = i.get_filename()[0]
                 i = os.path.join(self.get_target_dir(i), tmp)
+            elif isinstance(i, mesonlib.File):
+                i = os.path.join(i.subdir, i.fname)
+                if absolute_paths:
+                    i = os.path.join(self.environment.get_build_dir(), i)
+            # FIXME: str types are blindly added and ignore the 'absolute_paths' argument
+            elif not isinstance(i, str):
+                err_msg = 'Argument {0} is of unknown type {1}'
+                raise RuntimeError(err_msg.format(str(i), str(type(i))))
             for (j, src) in enumerate(srcs):
                 i = i.replace('@INPUT%d@' % j, src)
             for (j, res) in enumerate(ofilenames):
