@@ -131,6 +131,7 @@ class NinjaBackend(backends.Backend):
         self.ninja_filename = 'build.ninja'
         self.fortran_deps = {}
         self.all_outputs = {}
+        self.valgrind = environment.find_valgrind()
 
     def detect_vs_dep_prefix(self, outfile, tempfilename):
         '''VS writes its dependency in a locale dependent format.
@@ -339,10 +340,15 @@ int dummy;
             if not tname in self.processed_targets:
                 self.generate_target(t, outfile)
 
-    def generate_custom_target(self, target, outfile):
-        (srcs, ofilenames, cmd) = self.eval_custom_target_command(target)
+    def custom_target_generator_inputs(self, target, outfile):
+        for s in target.sources:
+            if hasattr(s, 'held_object'):
+                s = s.held_object
+            if isinstance(s, build.GeneratedList):
+                self.generate_genlist_for_target(s, target, outfile)
+
+    def unwrap_dep_list(self, target):
         deps = []
-        desc = 'Generating {0} with a {1} command.'
         for i in target.get_dependencies():
             # FIXME, should not grab element at zero but rather expand all.
             if isinstance(i, list):
@@ -351,6 +357,13 @@ int dummy;
             if isinstance(fname, list):
                 fname = fname[0]
             deps.append(os.path.join(self.get_target_dir(i), fname))
+        return deps
+
+    def generate_custom_target(self, target, outfile):
+        self.custom_target_generator_inputs(target, outfile)
+        (srcs, ofilenames, cmd) = self.eval_custom_target_command(target)
+        deps = self.unwrap_dep_list(target)
+        desc = 'Generating {0} with a {1} command.'
         if target.build_always:
             deps.append('PHONY')
         elem = NinjaBuildElement(self.all_outputs, ofilenames, 'CUSTOM_COMMAND', srcs)
@@ -388,19 +401,19 @@ int dummy;
 
     def generate_run_target(self, target, outfile):
         runnerscript = [sys.executable, self.environment.get_build_command(), '--internal', 'commandrunner']
-        deps = []
+        deps = self.unwrap_dep_list(target)
         arg_strings = []
         for i in target.args:
             if isinstance(i, str):
                 arg_strings.append(i)
             elif isinstance(i, (build.BuildTarget, build.CustomTarget)):
                 relfname = self.get_target_filename(i)
-                deps.append(relfname)
                 arg_strings.append(os.path.join(self.environment.get_build_dir(), relfname))
             else:
                 mlog.debug(str(i))
                 raise MesonException('Unreachable code in generate_run_target.')
         elem = NinjaBuildElement(self.all_outputs, target.name, 'CUSTOM_COMMAND', deps)
+        elem.add_dep(deps)
         cmd = runnerscript + [self.environment.get_source_dir(), self.environment.get_build_dir(), target.subdir]
         texe = target.command
         try:
@@ -416,6 +429,8 @@ int dummy;
                 if exe_wrap is not None:
                     cmd += [exe_wrap]
             cmd.append(abs_exe)
+        elif isinstance(texe, dependencies.ExternalProgram):
+            cmd += texe.get_command()
         else:
             cmd.append(target.command)
         cmd += arg_strings
@@ -481,18 +496,33 @@ int dummy;
         pickle.dump(d, ofile)
 
     def generate_target_install(self, d):
-        libdir = self.environment.get_libdir()
-        bindir = self.environment.get_bindir()
-
         should_strip = self.environment.coredata.get_builtin_option('strip')
         for t in self.build.get_targets().values():
             if t.should_install():
+                # Find the installation directory
                 outdir = t.get_custom_install_dir()
-                if outdir is None:
-                    if isinstance(t, build.Executable):
-                        outdir = bindir
-                    else:
-                        outdir = libdir
+                if outdir is not None:
+                    pass
+                elif isinstance(t, build.SharedLibrary):
+                    # For toolchains/platforms that need an import library for
+                    # linking (separate from the shared library with all the
+                    # code), we need to install the import library (dll.a/.lib)
+                    if t.get_import_filename():
+                        # Install the import library.
+                        i = [self.get_target_filename_for_linking(t),
+                             self.environment.get_import_lib_dir(),
+                             # It has no aliases, should not be stripped, and
+                             # doesn't have an install_rpath
+                             [], False, '']
+                        d.targets.append(i)
+                    outdir = self.environment.get_shared_lib_dir()
+                elif isinstance(t, build.StaticLibrary):
+                    outdir = self.environment.get_static_lib_dir()
+                elif isinstance(t, build.Executable):
+                    outdir = self.environment.get_bindir()
+                else:
+                    # XXX: Add BuildTarget-specific install dir cases here
+                    outdir = self.environment.get_libdir()
                 i = [self.get_target_filename(t), outdir, t.get_aliaslist(),\
                     should_strip, t.install_rpath]
                 d.targets.append(i)
@@ -569,15 +599,21 @@ int dummy;
                 visible_name = 'for top level tests'
             else:
                 visible_name = s
-            elem = NinjaBuildElement(self.all_outputs, 'test-' + s, 'CUSTOM_COMMAND', ['all', 'PHONY'])
+            elem = NinjaBuildElement(self.all_outputs, 'test:' + s, 'CUSTOM_COMMAND', ['all', 'PHONY'])
             elem.add_item('COMMAND', cmd + ['--suite=' + s])
             elem.add_item('DESC', 'Running test suite %s.' % visible_name)
             elem.add_item('pool', 'console')
             elem.write(outfile)
 
+            if self.valgrind:
+                velem = NinjaBuildElement(self.all_outputs, 'test-valgrind:' + s, 'CUSTOM_COMMAND', ['all', 'PHONY'])
+                velem.add_item('COMMAND', cmd + ['--wrapper=' + self.valgrind, '--suite=' + s])
+                velem.add_item('DESC', 'Running test suite %s under Valgrind.' % visible_name)
+                velem.add_item('pool', 'console')
+                velem.write(outfile)
+
     def generate_tests(self, outfile):
         (test_data, benchmark_data) = self.serialise_tests()
-        valgrind = environment.find_valgrind()
         script_root = self.environment.get_script_dir()
         cmd = [ sys.executable, self.environment.get_build_command(), '--internal', 'test' ]
         if not self.environment.coredata.get_builtin_option('stdsplit'):
@@ -592,9 +628,9 @@ int dummy;
         elem.write(outfile)
         self.write_test_suite_targets(cmd, outfile)
 
-        if valgrind:
+        if self.valgrind:
             velem = NinjaBuildElement(self.all_outputs, 'test-valgrind', 'CUSTOM_COMMAND', ['all', 'PHONY'])
-            velem.add_item('COMMAND', cmd + ['--wrapper=' + valgrind])
+            velem.add_item('COMMAND', cmd + ['--wrapper=' + self.valgrind])
             velem.add_item('DESC', 'Running test suite under Valgrind.')
             velem.add_item('pool', 'console')
             velem.write(outfile)
@@ -644,7 +680,7 @@ int dummy;
         outname_rel = os.path.join(self.get_target_dir(target), fname)
         src_list = target.get_sources()
         class_list = []
-        compiler = self.get_compiler_for_source(src_list[0])
+        compiler = self.get_compiler_for_source(src_list[0], False)
         assert(compiler.get_language() == 'java')
         c = 'c'
         m = ''
@@ -695,7 +731,7 @@ int dummy;
         fname = target.get_filename()
         outname_rel = os.path.join(self.get_target_dir(target), fname)
         src_list = target.get_sources()
-        compiler = self.get_compiler_for_source(src_list[0])
+        compiler = self.get_compiler_for_source(src_list[0], False)
         assert(compiler.get_language() == 'cs')
         rel_srcs = [s.rel_to_builddir(self.build_to_src) for s in src_list]
         deps = []
@@ -765,7 +801,7 @@ int dummy;
                 if hasattr(i, 'fname'):
                     i = i.fname
                 if i.endswith('vala'):
-                    vapiname = os.path.splitext(os.path.split(i)[1])[0] + '.vapi'
+                    vapiname = dep.name + '.vapi'
                     fullname = os.path.join(self.get_target_dir(dep), vapiname)
                     result.append(fullname)
                     break
@@ -783,7 +819,7 @@ int dummy;
                 vala_input_files.append(s.rel_to_builddir(self.build_to_src))
         if len(src) == 0:
             raise InvalidArguments('Vala library has no Vala source files.')
-        namebase = os.path.splitext(os.path.split(src[0].fname)[1])[0]
+        namebase = target.name
         base_h = namebase + '.h'
         base_vapi = namebase + '.vapi'
         hname = os.path.normpath(os.path.join(self.get_target_dir(target), base_h))
@@ -1213,7 +1249,7 @@ rule FORTRAN_DEP_HACK
         else:
             command_template = ' command = %s %s $ARGS %s %s %s $in\n'
         command = command_template % \
-            (' '.join(compiler.get_exelist()),\
+            (' '.join([ninja_quote(i) for i in compiler.get_exelist()]),\
              ' '.join(cross_args),
              ' '.join(quoted_depargs),\
              ' '.join(compiler.get_output_args('$out')),\
@@ -1296,40 +1332,43 @@ rule FORTRAN_DEP_HACK
         for genlist in target.get_generated_sources():
             if isinstance(genlist, build.CustomTarget):
                 continue # Customtarget has already written its output rules
-            generator = genlist.get_generator()
-            exe = generator.get_exe()
-            exe_arr = self.exe_object_to_cmd_array(exe)
-            infilelist = genlist.get_infilelist()
-            outfilelist = genlist.get_outfilelist()
-            base_args = generator.get_arglist()
-            extra_dependencies = [os.path.join(self.build_to_src, i) for i in genlist.extra_depends]
-            for i in range(len(infilelist)):
-                if len(generator.outputs) == 1:
-                    sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
-                else:
-                    sole_output = ''
-                curfile = infilelist[i]
-                infilename = os.path.join(self.build_to_src, curfile)
-                outfiles = genlist.get_outputs_for(curfile)
-                outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
-                args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
-                        for x in base_args]
-                args = self.replace_outputs(args, self.get_target_private_dir(target), outfilelist)
-                # We have consumed output files, so drop them from the list of remaining outputs.
-                if sole_output == '':
-                    outfilelist = outfilelist[len(generator.outputs):]
-                relout = self.get_target_private_dir(target)
-                args = [x.replace("@SOURCE_DIR@", self.build_to_src).replace("@BUILD_DIR@", relout)
-                        for x in args]
-                cmdlist = exe_arr + self.replace_extra_args(args, genlist)
-                elem = NinjaBuildElement(self.all_outputs, outfiles, 'CUSTOM_COMMAND', infilename)
-                if len(extra_dependencies) > 0:
-                    elem.add_dep(extra_dependencies)
-                elem.add_item('DESC', 'Generating $out')
-                if isinstance(exe, build.BuildTarget):
-                    elem.add_dep(self.get_target_filename(exe))
-                elem.add_item('COMMAND', cmdlist)
-                elem.write(outfile)
+            self.generate_genlist_for_target(genlist, target, outfile)
+
+    def generate_genlist_for_target(self, genlist, target, outfile):
+        generator = genlist.get_generator()
+        exe = generator.get_exe()
+        exe_arr = self.exe_object_to_cmd_array(exe)
+        infilelist = genlist.get_infilelist()
+        outfilelist = genlist.get_outfilelist()
+        base_args = generator.get_arglist()
+        extra_dependencies = [os.path.join(self.build_to_src, i) for i in genlist.extra_depends]
+        for i in range(len(infilelist)):
+            if len(generator.outputs) == 1:
+                sole_output = os.path.join(self.get_target_private_dir(target), outfilelist[i])
+            else:
+                sole_output = ''
+            curfile = infilelist[i]
+            infilename = os.path.join(self.build_to_src, curfile)
+            outfiles = genlist.get_outputs_for(curfile)
+            outfiles = [os.path.join(self.get_target_private_dir(target), of) for of in outfiles]
+            args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
+                    for x in base_args]
+            args = self.replace_outputs(args, self.get_target_private_dir(target), outfilelist)
+            # We have consumed output files, so drop them from the list of remaining outputs.
+            if sole_output == '':
+                outfilelist = outfilelist[len(generator.outputs):]
+            relout = self.get_target_private_dir(target)
+            args = [x.replace("@SOURCE_DIR@", self.build_to_src).replace("@BUILD_DIR@", relout)
+                    for x in args]
+            cmdlist = exe_arr + self.replace_extra_args(args, genlist)
+            elem = NinjaBuildElement(self.all_outputs, outfiles, 'CUSTOM_COMMAND', infilename)
+            if len(extra_dependencies) > 0:
+                elem.add_dep(extra_dependencies)
+            elem.add_item('DESC', 'Generating $out')
+            if isinstance(exe, build.BuildTarget):
+                elem.add_dep(self.get_target_filename(exe))
+            elem.add_item('COMMAND', cmdlist)
+            elem.write(outfile)
 
     def scan_fortran_module_outputs(self, target):
         compiler = None
@@ -1400,25 +1439,25 @@ rule FORTRAN_DEP_HACK
         if isinstance(src, RawFilename) and src.fname.endswith('.h'):
             raise RuntimeError('Fug')
         extra_orderdeps = []
-        compiler = self.get_compiler_for_source(src)
+        compiler = self.get_compiler_for_source(src, target.is_cross)
         commands = []
         # The first thing is implicit include directories: source, build and private.
         commands += compiler.get_include_args(self.get_target_private_dir(target), False)
+        # Compiler args for compiling this target
         commands += compilers.get_base_compile_args(self.environment.coredata.base_options,
                                                     compiler)
+        # Add the root source and build directories as include dirs
         curdir = target.get_subdir()
         tmppath = os.path.normpath(os.path.join(self.build_to_src, curdir))
         commands += compiler.get_include_args(tmppath, False)
         if curdir ==  '':
             curdir = '.'
         commands += compiler.get_include_args(curdir, False)
-        commands += self.generate_basic_compiler_args(target, compiler)
-        # -I args work differently than other ones. In them the
-        # first found directory is used whereas for other flags
-        # (such as -ffoo -fno-foo) the latest one is used.
-        # Therefore put the internal include directories here
-        # at the beginning so they override args coming from
-        # e.g. pkg-config.
+        # -I args work differently than other ones. In them the first found
+        # directory is used whereas for other flags (such as -ffoo -fno-foo) the
+        # latest one is used.  Therefore put the internal include directories
+        # here before generating the "basic compiler args" so they override args
+        # coming from e.g. pkg-config.
         for i in target.get_include_dirs():
             basedir = i.get_curdir()
             for d in i.get_incdirs():
@@ -1430,6 +1469,7 @@ rule FORTRAN_DEP_HACK
                 commands += sargs
             for d in i.get_extra_build_dirs():
                 commands += compiler.get_include_args(d, i.is_system)
+        commands += self.generate_basic_compiler_args(target, compiler)
         for d in target.external_deps:
             if d.need_threads():
                 commands += compiler.thread_flags()
@@ -1640,8 +1680,12 @@ rule FORTRAN_DEP_HACK
             else:
                 soversion = None
             commands += linker.get_soname_args(target.name, abspath, soversion)
+            # This is only visited when using the Visual Studio toolchain
             if target.vs_module_defs and hasattr(linker, 'gen_vs_module_defs_args'):
                 commands += linker.gen_vs_module_defs_args(target.vs_module_defs.rel_to_builddir(self.build_to_src))
+            # This is only visited when building for Windows using either MinGW/GCC or Visual Studio
+            if target.import_filename:
+                commands += linker.gen_import_library_args(os.path.join(target.subdir, target.import_filename))
         elif isinstance(target, build.StaticLibrary):
             commands += linker.get_std_link_args()
         else:
@@ -1659,8 +1703,7 @@ rule FORTRAN_DEP_HACK
                 commands += linker.thread_link_flags()
         if not isinstance(target, build.StaticLibrary):
             commands += target.link_args
-        # External deps must be last because target link libraries may depend on them.
-        if not(isinstance(target, build.StaticLibrary)):
+            # External deps must be last because target link libraries may depend on them.
             for dep in target.get_external_deps():
                 commands += dep.get_link_args()
             for d in target.get_dependencies():
@@ -1778,7 +1821,7 @@ rule FORTRAN_DEP_HACK
 
         ninja_command = environment.detect_ninja()
         if ninja_command is None:
-            raise MesonException('Could not detect ninja command')
+            raise MesonException('Could not detect Ninja v1.6 or newer)')
         elem = NinjaBuildElement(self.all_outputs, 'clean', 'CUSTOM_COMMAND', 'PHONY')
         elem.add_item('COMMAND', [ninja_command, '-t', 'clean'])
         elem.add_item('description', 'Cleaning')
