@@ -59,12 +59,15 @@ gnulike_buildtype_args = {'plain' : [],
                           # See https://github.com/mesonbuild/meson/pull/509
                           'debug' : ['-O0', '-g'],
                           'debugoptimized' : ['-O2', '-g'],
-                          'release' : ['-O3']}
+                          'release' : ['-O3'],
+                          'minsize' : ['-Os', '-g']}
 
 msvc_buildtype_args = {'plain' : [],
                        'debug' : ["/MDd", "/ZI", "/Ob0", "/Od", "/RTC1"],
                        'debugoptimized' : ["/MD", "/Zi", "/O2", "/Ob1"],
-                       'release' : ["/MD", "/O2", "/Ob2"]}
+                       'release' : ["/MD", "/O2", "/Ob2"],
+                       'minsize' : ["/MD", "/Zi", "/Os", "/Ob1"],
+                       }
 
 gnulike_buildtype_linker_args = {}
 
@@ -74,38 +77,50 @@ if mesonlib.is_osx():
                                           'debug' : [],
                                           'debugoptimized' : [],
                                           'release' : [],
+                                          'minsize' : [],
                                          })
 else:
     gnulike_buildtype_linker_args.update({'plain' : [],
                                           'debug' : [],
                                           'debugoptimized' : [],
                                           'release' : ['-Wl,-O1'],
+                                          'minsize' : [],
                                          })
 
 msvc_buildtype_linker_args = {'plain' : [],
                               'debug' : [],
                               'debugoptimized' : [],
-                              'release' : []}
+                              'release' : [],
+                              'minsize' : ['/INCREMENTAL:NO'],
+                              }
 
 java_buildtype_args = {'plain' : [],
                        'debug' : ['-g'],
                        'debugoptimized' : ['-g'],
-                       'release' : []}
+                       'release' : [],
+                       'minsize' : [],
+                       }
 
 rust_buildtype_args = {'plain' : [],
                        'debug' : ['-g'],
                        'debugoptimized' : ['-g', '--opt-level', '2'],
-                       'release' : ['--opt-level', '3']}
+                       'release' : ['--opt-level', '3'],
+                       'minsize' : [],
+                       }
 
 mono_buildtype_args = {'plain' : [],
                        'debug' : ['-debug'],
                        'debugoptimized': ['-debug', '-optimize+'],
-                       'release' : ['-optimize+']}
+                       'release' : ['-optimize+'],
+                       'minsize' : [],
+                       }
 
 swift_buildtype_args = {'plain' : [],
                         'debug' : ['-g'],
                         'debugoptimized': ['-g', '-O'],
-                        'release' : ['-O']}
+                        'release' : ['-O'],
+                        'minsize' : [],
+                        }
 
 gnu_winlibs = ['-lkernel32', '-luser32', '-lgdi32', '-lwinspool', '-lshell32',
                '-lole32', '-loleaut32', '-luuid', '-lcomdlg32', '-ladvapi32']
@@ -114,6 +129,10 @@ msvc_winlibs = ['kernel32.lib', 'user32.lib', 'gdi32.lib',
                 'winspool.lib', 'shell32.lib', 'ole32.lib', 'oleaut32.lib',
                 'uuid.lib', 'comdlg32.lib', 'advapi32.lib']
 
+gnu_color_args = {'auto' : ['-fdiagnostics-color=auto'],
+                  'always': ['-fdiagnostics-color=always'],
+                  'never' : ['-fdiagnostics-color=never'],
+                  }
 
 base_options = {
                 'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled headers', True),
@@ -129,6 +148,9 @@ base_options = {
                 'b_coverage': coredata.UserBooleanOption('b_coverage',
                                                          'Enable coverage tracking.',
                                                          False),
+                'b_colorout' : coredata.UserComboOption('b_colorout', 'Use colored output',
+                                                        ['auto', 'always', 'never'], 
+                                                        'always'),
                 }
 
 def sanitizer_compile_args(value):
@@ -153,6 +175,10 @@ def get_base_compile_args(options, compiler):
             args.append('-flto')
     except KeyError:
         pass
+    try:
+        args += compiler.get_colorout_args(options['b_colorout'].value)
+    except KeyError:
+        pass
     try:
         args += sanitizer_compile_args(options['b_sanitize'].value)
     except KeyError:
@@ -248,6 +274,13 @@ class Compiler():
     def get_linker_always_args(self):
         return []
 
+    def gen_import_library_args(self, implibname):
+        """
+        Used only on Windows for libraries that need an import library.
+        This currently means C, C++, Fortran.
+        """
+        return []
+
     def get_options(self):
         return {} # build afresh every time
 
@@ -282,17 +315,37 @@ class Compiler():
         raise EnvironmentException('Language %s does not support function checks.' % self.language)
 
     def unix_link_flags_to_native(self, args):
-        return args
+        "Always returns a copy that can be independently mutated"
+        return args[:]
 
     def unix_compile_flags_to_native(self, args):
-        return args
+        "Always returns a copy that can be independently mutated"
+        return args[:]
 
-    def find_library(self, libname, extra_dirs):
+    def find_library(self, *args, **kwargs):
         raise EnvironmentException('Language {} does not support library finding.'.format(self.language))
 
     def get_library_dirs(self):
         return []
 
+    def has_argument(self, arg):
+        raise EnvironmentException('Language {} does not support has_arg.'.format(self.language))
+
+    def get_cross_extra_flags(self, environment, *, compile, link):
+        extra_flags = []
+        if self.is_cross and environment:
+            if 'properties' in environment.cross_info.config:
+                lang_args_key = self.language + '_args'
+                if compile:
+                    extra_flags += environment.cross_info.config['properties'].get(lang_args_key, [])
+                lang_link_args_key = self.language + '_link_args'
+                if link:
+                    extra_flags += environment.cross_info.config['properties'].get(lang_link_args_key, [])
+        return extra_flags
+
+    def get_colorout_args(self, colortype):
+        return []
+
 class CCompiler(Compiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version)
@@ -311,6 +364,13 @@ class CCompiler(Compiler):
     def get_always_args(self):
         return []
 
+    def get_linker_debug_crt_args(self):
+        """
+        Arguments needed to select a debug crt for the linker
+        This is only needed for MSVC
+        """
+        return []
+
     def get_no_stdinc_args(self):
         return ['-nostdinc']
 
@@ -420,7 +480,15 @@ class CCompiler(Compiler):
     def get_linker_search_args(self, dirname):
         return ['-L'+dirname]
 
-    def sanity_check_impl(self, work_dir, sname, code):
+    def gen_import_library_args(self, implibname):
+        """
+        The name of the outputted import library
+
+        This implementation is used only on Windows by compilers that use GNU ld
+        """
+        return ['-Wl,--out-implib=' + implibname]
+
+    def sanity_check_impl(self, work_dir, environment, sname, code):
         mlog.debug('Sanity testing ' + self.language + ' compiler:', ' '.join(self.exelist))
         mlog.debug('Is cross compiler: %s.' % str(self.is_cross))
 
@@ -435,7 +503,10 @@ class CCompiler(Compiler):
                 # on OSX the compiler binary is the same but you need
                 # a ton of compiler flags to differentiate between
                 # arm and x86_64. So just compile.
-                extra_flags = self.get_compile_only_args()
+                extra_flags += self.get_cross_extra_flags(environment, compile=True, link=False)
+                extra_flags += self.get_compile_only_args()
+            else:
+                extra_flags += self.get_cross_extra_flags(environment, compile=True, link=True)
         # Is a valid executable output for all toolchains and platforms
         binname += '.exe'
         # Write binary check source
@@ -471,30 +542,36 @@ class CCompiler(Compiler):
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by {0} compiler {1} are not runnable.'.format(self.language, self.name_string()))
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         code = 'int main(int argc, char **argv) { int class=0; return class; }\n'
-        return self.sanity_check_impl(work_dir, 'sanitycheckc.c', code)
+        return self.sanity_check_impl(work_dir, environment, 'sanitycheckc.c', code)
 
-    def has_header(self, hname, extra_args=[]):
+    def has_header(self, hname, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         templ = '''#include<%s>
 int someSymbolHereJustForFun;
 '''
-        return self.compiles(templ % hname, extra_args)
+        return self.compiles(templ % hname, env, extra_args)
 
-    def has_header_symbol(self, hname, symbol, prefix, extra_args=[]):
+    def has_header_symbol(self, hname, symbol, prefix, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         templ = '''{2}
 #include <{0}>
 int main () {{ {1}; }}'''
         # Pass -O0 to ensure that the symbol isn't optimized away
-        extra_args += self.get_no_optimization_args()
-        return self.compiles(templ.format(hname, symbol, prefix), extra_args)
+        args = extra_args + self.get_no_optimization_args()
+        return self.compiles(templ.format(hname, symbol, prefix), env, args)
 
-    def compile(self, code, srcname, extra_args=[]):
+    def compile(self, code, srcname, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         commands = self.get_exelist()
         commands.append(srcname)
         commands += extra_args
         mlog.debug('Running compile:')
-        mlog.debug('Command line: ', ' '.join(commands))
+        mlog.debug('Command line: ', ' '.join(commands), '\n')
         mlog.debug('Code:\n', code)
         p = subprocess.Popen(commands, cwd=os.path.split(srcname)[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         (stde, stdo) = p.communicate()
@@ -505,15 +582,24 @@ int main () {{ {1}; }}'''
         os.remove(srcname)
         return p
 
-    def compiles(self, code, extra_args = []):
+    def compiles(self, code, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
+        if isinstance(extra_args, str):
+            extra_args = [extra_args]
         suflen = len(self.default_suffix)
         (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
         os.close(fd)
         ofile = open(srcname, 'w')
         ofile.write(code)
         ofile.close()
-        extra_args = extra_args + self.get_compile_only_args()
-        p = self.compile(code, srcname, extra_args)
+        # Convert flags to the native type of the selected compiler
+        args = self.unix_link_flags_to_native(extra_args)
+        # Read c_args/cpp_args/etc from the cross-info file (if needed)
+        args += self.get_cross_extra_flags(env, compile=True, link=False)
+        # We only want to compile; not link
+        args += self.get_compile_only_args()
+        p = self.compile(code, srcname, args)
         try:
             trial = srcname[:-suflen] + 'o'
             os.remove(trial)
@@ -525,7 +611,11 @@ int main () {{ {1}; }}'''
             pass
         return p.returncode == 0
 
-    def links(self, code, extra_args = []):
+    def links(self, code, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
+        elif isinstance(extra_args, str):
+            extra_args = [extra_args]
         (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
         os.close(fd)
         (fd, dstname) = tempfile.mkstemp()
@@ -533,22 +623,24 @@ int main () {{ {1}; }}'''
         ofile = open(srcname, 'w')
         ofile.write(code)
         ofile.close()
-        # Need to add buildtype args to select the CRT to use with MSVC
-        # This is needed especially while trying to link with static libraries
-        # since MSVC won't auto-select a CRT for us in that case and will error
-        # out asking us to select one.
-        extra_args = self.get_buildtype_args('debug') + \
-            self.unix_link_flags_to_native(extra_args) + \
-            self.get_output_args(dstname)
-        p = self.compile(code, srcname, extra_args)
+        # Convert flags to the native type of the selected compiler
+        args = self.unix_link_flags_to_native(extra_args)
+        # Select a CRT if needed since we're linking
+        args += self.get_linker_debug_crt_args()
+        # Read c_args/c_link_args/cpp_args/cpp_link_args/etc from the cross-info file (if needed)
+        args += self.get_cross_extra_flags(env, compile=True, link=True)
+        # Arguments specifying the output filename
+        args += self.get_output_args(dstname)
+        p = self.compile(code, srcname, args)
         try:
             os.remove(dstname)
         except FileNotFoundError:
             pass
         return p.returncode == 0
 
-    def run(self, code, extra_args=[]):
-        mlog.debug('Running code:\n\n', code)
+    def run(self, code, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         if self.is_cross and self.exe_wrapper is None:
             raise CrossNoRunException('Can not run test applications in this cross environment.')
         (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
@@ -556,19 +648,27 @@ int main () {{ {1}; }}'''
         ofile = open(srcname, 'w')
         ofile.write(code)
         ofile.close()
+        # Convert flags to the native type of the selected compiler
+        args = self.unix_link_flags_to_native(extra_args)
+        # Select a CRT if needed since we're linking
+        args += self.get_linker_debug_crt_args()
+        # Read c_link_args/cpp_link_args/etc from the cross-info file
+        args += self.get_cross_extra_flags(env, compile=True, link=True)
+        # Create command list
         exename = srcname + '.exe' # Is guaranteed to be executable on every platform.
-        commands = self.get_exelist()
-        # Same reasoning as self.links() above
-        commands += self.get_buildtype_args('debug')
-        commands += self.unix_link_flags_to_native(extra_args)
+        commands = self.get_exelist() + args
         commands.append(srcname)
         commands += self.get_output_args(exename)
+        mlog.debug('Running code:\n\n', code)
+        mlog.debug('Command line:', ' '.join(commands))
         p = subprocess.Popen(commands, cwd=os.path.split(srcname)[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         (stdo, stde) = p.communicate()
         stde = stde.decode()
         stdo = stdo.decode()
-        mlog.debug('Compiler stdout:\n', stdo)
-        mlog.debug('Compiler stderr:\n', stde)
+        mlog.debug('Compiler stdout:\n')
+        mlog.debug(stdo)
+        mlog.debug('Compiler stderr:\n')
+        mlog.debug(stde)
         os.remove(srcname)
         if p.returncode != 0:
             return RunResult(False)
@@ -585,8 +685,10 @@ int main () {{ {1}; }}'''
         (so, se) = pe.communicate()
         so = so.decode()
         se = se.decode()
-        mlog.debug('Program stdout:\n', so)
-        mlog.debug('Program stderr:\n', se)
+        mlog.debug('Program stdout:\n')
+        mlog.debug(so)
+        mlog.debug('Program stderr:\n')
+        mlog.debug(se)
         try:
             os.remove(exename)
         except PermissionError:
@@ -596,7 +698,9 @@ int main () {{ {1}; }}'''
             pass
         return RunResult(True, pe.returncode, so, se)
 
-    def cross_sizeof(self, element, prefix, env, extra_args=[]):
+    def cross_sizeof(self, element, prefix, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         element_exists_templ = '''#include <stdio.h>
 {0}
 int main(int argc, char **argv) {{
@@ -607,16 +711,12 @@ int main(int argc, char **argv) {{
 %s
 int temparray[%d-sizeof(%s)];
 '''
-        try:
-            extra_args += env.cross_info.config['properties'][self.language + '_args']
-        except KeyError:
-            pass
-        extra_args += self.get_no_optimization_args()
-        if not self.compiles(element_exists_templ.format(prefix, element)):
+        args = extra_args + self.get_no_optimization_args()
+        if not self.compiles(element_exists_templ.format(prefix, element), env, args):
             return -1
         for i in range(1, 1024):
             code = templ % (prefix, i, element)
-            if self.compiles(code, extra_args):
+            if self.compiles(code, env, args):
                 if self.id == 'msvc':
                     # MSVC refuses to construct an array of zero size, so
                     # the test only succeeds when i is sizeof(element) + 1
@@ -624,7 +724,9 @@ int temparray[%d-sizeof(%s)];
                 return i
         raise EnvironmentException('Cross checking sizeof overflowed.')
 
-    def sizeof(self, element, prefix, env, extra_args=[]):
+    def sizeof(self, element, prefix, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         if self.is_cross:
             return self.cross_sizeof(element, prefix, env, extra_args)
         templ = '''#include<stdio.h>
@@ -635,14 +737,16 @@ int main(int argc, char **argv) {
     return 0;
 };
 '''
-        res = self.run(templ % (prefix, element), extra_args)
+        res = self.run(templ % (prefix, element), env, extra_args)
         if not res.compiled:
             return -1
         if res.returncode != 0:
             raise EnvironmentException('Could not run sizeof test binary.')
         return int(res.stdout)
 
-    def cross_alignment(self, typename, env, extra_args=[]):
+    def cross_alignment(self, typename, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         type_exists_templ = '''#include <stdio.h>
 int main(int argc, char **argv) {{
     {0} something;
@@ -656,16 +760,12 @@ struct tmp {
 
 int testarray[%d-offsetof(struct tmp, target)];
 '''
-        try:
-            extra_args += env.cross_info.config['properties'][self.language + '_args']
-        except KeyError:
-            pass
-        extra_args += self.get_no_optimization_args()
-        if not self.compiles(type_exists_templ.format(typename)):
+        args = extra_args + self.get_no_optimization_args()
+        if not self.compiles(type_exists_templ.format(typename), env, args):
             return -1
         for i in range(1, 1024):
             code = templ % (typename, i)
-            if self.compiles(code, extra_args):
+            if self.compiles(code, env, args):
                 if self.id == 'msvc':
                     # MSVC refuses to construct an array of zero size, so
                     # the test only succeeds when i is sizeof(element) + 1
@@ -673,7 +773,9 @@ int testarray[%d-offsetof(struct tmp, target)];
                 return i
         raise EnvironmentException('Cross checking offsetof overflowed.')
 
-    def alignment(self, typename, env, extra_args=[]):
+    def alignment(self, typename, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         if self.is_cross:
             return self.cross_alignment(typename, env, extra_args)
         templ = '''#include<stdio.h>
@@ -689,7 +791,7 @@ int main(int argc, char **argv) {
   return 0;
 }
 '''
-        res = self.run(templ % typename, extra_args)
+        res = self.run(templ % typename, env, extra_args)
         if not res.compiled:
             raise EnvironmentException('Could not compile alignment test.')
         if res.returncode != 0:
@@ -699,7 +801,7 @@ int main(int argc, char **argv) {
             raise EnvironmentException('Could not determine alignment of %s. Sorry. You might want to file a bug.' % typename)
         return align
 
-    def has_function(self, funcname, prefix, env, extra_args=[]):
+    def has_function(self, funcname, prefix, env, extra_args=None):
         """
         First, this function looks for the symbol in the default libraries
         provided by the compiler (stdlib + a few others usually). If that
@@ -707,6 +809,8 @@ int main(int argc, char **argv) {
         an implementation of the function, and if that fails, it checks if it's
         implemented as a compiler-builtin.
         """
+        if extra_args is None:
+            extra_args = []
         # Define the symbol to something else in case it is defined by the
         # includes or defines listed by the user `{0}` or by the compiler.
         # Then, undef the symbol to get rid of it completely.
@@ -754,51 +858,55 @@ int main(int argc, char **argv) {
                 if isinstance(val, bool):
                     return val
                 raise EnvironmentException('Cross variable {0} is not a boolean.'.format(varname))
-        if self.links(templ.format(prefix, funcname), extra_args):
+        if self.links(templ.format(prefix, funcname), env, extra_args):
             return True
         # Add -O0 to ensure that the symbol isn't optimized away by the compiler
-        extra_args += self.get_no_optimization_args()
+        args = extra_args + self.get_no_optimization_args()
         # Sometimes the implementation is provided by the header, or the header
         # redefines the symbol to be something else. In that case, we want to
         # still detect the function. We still want to fail if __stub_foo or
         # _stub_foo are defined, of course.
         header_templ = '#include <limits.h>\n{0}\n' + stubs_fail + '\nint main() {{ {1}; }}'
-        if self.links(header_templ.format(prefix, funcname), extra_args):
+        if self.links(header_templ.format(prefix, funcname), env, args):
             return True
         # Some functions like alloca() are defined as compiler built-ins which
         # are inlined by the compiler, so test for that instead. Built-ins are
         # special functions that ignore all includes and defines, so we just
         # directly try to link via main().
-        return self.links('int main() {{ {0}; }}'.format('__builtin_' + funcname), extra_args)
+        return self.links('int main() {{ {0}; }}'.format('__builtin_' + funcname), env, args)
 
-    def has_member(self, typename, membername, prefix, extra_args=[]):
+    def has_member(self, typename, membername, prefix, env, extra_args=None):
+        if extra_args is None:
+            extra_args = []
         templ = '''%s
 void bar() {
     %s foo;
     foo.%s;
 };
 '''
-        return self.compiles(templ % (prefix, typename, membername), extra_args)
+        return self.compiles(templ % (prefix, typename, membername), env, extra_args)
 
-    def has_type(self, typename, prefix, extra_args):
+    def has_type(self, typename, prefix, env, extra_args):
         templ = '''%s
 void bar() {
     sizeof(%s);
 };
 '''
-        return self.compiles(templ % (prefix, typename), extra_args)
+        return self.compiles(templ % (prefix, typename), env, extra_args)
 
-    def find_library(self, libname, extra_dirs):
+    def find_library(self, libname, env, extra_dirs):
         # First try if we can just add the library as -l.
         code = '''int main(int argc, char **argv) {
     return 0;
 }
         '''
+        if extra_dirs and isinstance(extra_dirs, str):
+            extra_dirs = [extra_dirs]
         # Gcc + co seem to prefer builtin lib dirs to -L dirs.
         # Only try to find std libs if no extra dirs specified.
         if len(extra_dirs) == 0:
             args = ['-l' + libname]
-            if self.links(code, extra_args=args):
+            if self.links(code, env, extra_args=args):
                 return args
         # Not found? Try to find the library file itself.
         extra_dirs += self.get_library_dirs()
@@ -819,6 +927,9 @@ void bar() {
     def thread_link_flags(self):
         return ['-pthread']
 
+    def has_argument(self, arg, env):
+        return self.compiles('int i;\n', env, extra_args=arg)
+
 class CPPCompiler(CCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrap):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
@@ -831,9 +942,9 @@ class CPPCompiler(CCompiler):
             return True
         return False
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         code = 'class breakCCompiler;int main(int argc, char **argv) { return 0; }\n'
-        return self.sanity_check_impl(work_dir, 'sanitycheckcpp.cc', code)
+        return self.sanity_check_impl(work_dir, environment, 'sanitycheckcpp.cc', code)
 
 class ObjCCompiler(CCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrap):
@@ -847,16 +958,23 @@ class ObjCCompiler(CCompiler):
             return True
         return False
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
+        # TODO try to use sanity_check_impl instead of duplicated code
         source_name = os.path.join(work_dir, 'sanitycheckobjc.m')
         binary_name = os.path.join(work_dir, 'sanitycheckobjc')
+        extra_flags = self.get_cross_extra_flags(environment, compile=True, link=False)
+        if self.is_cross:
+            extra_flags += self.get_compile_only_args()
         ofile = open(source_name, 'w')
         ofile.write('#import<stdio.h>\nint main(int argc, char **argv) { return 0; }\n')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
+        pc = subprocess.Popen(self.exelist + extra_flags + [source_name, '-o', binary_name])
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('ObjC compiler %s can not compile programs.' % self.name_string())
+        if self.is_cross:
+            # Can't check if the binaries run so we have to assume they do
+            return
         pe = subprocess.Popen(binary_name)
         pe.wait()
         if pe.returncode != 0:
@@ -874,16 +992,23 @@ class ObjCPPCompiler(CPPCompiler):
             return True
         return False
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
+        # TODO try to use sanity_check_impl instead of duplicated code
         source_name = os.path.join(work_dir, 'sanitycheckobjcpp.mm')
         binary_name = os.path.join(work_dir, 'sanitycheckobjcpp')
+        extra_flags = self.get_cross_extra_flags(environment, compile=True, link=False)
+        if self.is_cross:
+            extra_flags += self.get_compile_only_args()
         ofile = open(source_name, 'w')
         ofile.write('#import<stdio.h>\nclass MyClass;int main(int argc, char **argv) { return 0; }\n')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
+        pc = subprocess.Popen(self.exelist + extra_flags + [source_name, '-o', binary_name])
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('ObjC++ compiler %s can not compile programs.' % self.name_string())
+        if self.is_cross:
+            # Can't check if the binaries run so we have to assume they do
+            return
         pe = subprocess.Popen(binary_name)
         pe.wait()
         if pe.returncode != 0:
@@ -972,7 +1097,7 @@ class MonoCompiler(Compiler):
     def get_pch_name(self, header_name):
         return ''
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         src = 'sanity.cs'
         obj = 'sanity.exe'
         source_name = os.path.join(work_dir, src)
@@ -1084,7 +1209,7 @@ class JavaCompiler(Compiler):
     def get_buildtype_args(self, buildtype):
         return java_buildtype_args[buildtype]
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         src = 'SanityCheck.java'
         obj = 'SanityCheck'
         source_name = os.path.join(work_dir, src)
@@ -1132,7 +1257,7 @@ class ValaCompiler(Compiler):
     def get_language(self):
         return self.language
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         src = 'valatest.vala'
         source_name = os.path.join(work_dir, src)
         ofile = open(source_name, 'w')
@@ -1140,7 +1265,8 @@ class ValaCompiler(Compiler):
 }
 ''')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + ['-C', '-c', src], cwd=work_dir)
+        extra_flags = self.get_cross_extra_flags(environment, compile=True, link=False)
+        pc = subprocess.Popen(self.exelist + extra_flags + ['-C', '-c', src], cwd=work_dir)
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Vala compiler %s can not compile programs.' % self.name_string())
@@ -1150,7 +1276,7 @@ class ValaCompiler(Compiler):
         return suffix in ('vala', 'vapi')
 
     def get_buildtype_args(self, buildtype):
-        if buildtype == 'debug' or buildtype == 'debugoptimized':
+        if buildtype == 'debug' or buildtype == 'debugoptimized' or buildtype == 'minsize':
             return ['--debug']
         return []
 
@@ -1175,7 +1301,7 @@ class RustCompiler(Compiler):
     def get_language(self):
         return self.language
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         source_name = os.path.join(work_dir, 'sanity.rs')
         output_name = os.path.join(work_dir, 'rusttest')
         ofile = open(source_name, 'w')
@@ -1273,7 +1399,7 @@ class SwiftCompiler(Compiler):
     def get_compile_only_args(self):
         return ['-c']
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         src = 'swifttest.swift'
         source_name = os.path.join(work_dir, src)
         output_name = os.path.join(work_dir, 'swifttest')
@@ -1281,7 +1407,8 @@ class SwiftCompiler(Compiler):
         ofile.write('''1 + 2
 ''')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + ['-emit-executable', '-o', output_name, src], cwd=work_dir)
+        extra_flags = self.get_cross_extra_flags(environment, compile=True, link=True)
+        pc = subprocess.Popen(self.exelist + extra_flags + ['-emit-executable', '-o', output_name, src], cwd=work_dir)
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Swift compiler %s can not compile programs.' % self.name_string())
@@ -1313,6 +1440,17 @@ class VisualStudioCCompiler(CCompiler):
     def get_always_args(self):
         return self.always_args
 
+    def get_linker_debug_crt_args(self):
+        """
+        Arguments needed to select a debug crt for the linker
+
+        Sometimes we need to manually select the CRT (C runtime) to use with
+        MSVC. One example is when trying to link with static libraries since
+        MSVC won't auto-select a CRT for us in that case and will error out
+        asking us to select one.
+        """
+        return ['/MDd']
+
     def get_buildtype_args(self, buildtype):
         return msvc_buildtype_args[buildtype]
 
@@ -1376,6 +1514,10 @@ class VisualStudioCCompiler(CCompiler):
         objname = os.path.splitext(pchname)[0] + '.obj'
         return (objname, ['/Yc' + header, '/Fp' + pchname, '/Fo' + objname ])
 
+    def gen_import_library_args(self, implibname):
+        "The name of the outputted import library"
+        return ['/IMPLIB:' + implibname]
+
     def build_rpath_args(self, build_dir, rpath_paths, install_rpath):
         return []
 
@@ -1401,7 +1543,7 @@ class VisualStudioCCompiler(CCompiler):
             if i.startswith('-L'):
                 i = '/LIBPATH:' + i[2:]
             # Translate GNU-style -lfoo library name to the import library
-            if i.startswith('-l'):
+            elif i.startswith('-l'):
                 name = i[2:]
                 if name in ('m', 'c'):
                     # With MSVC, these are provided by the C runtime which is
@@ -1427,6 +1569,30 @@ class VisualStudioCCompiler(CCompiler):
         # msvc does not have a concept of system header dirs.
         return ['-I' + path]
 
+    # Visual Studio is special. It ignores arguments it does not
+    # understand and you can't tell it to error out on those.
+    # http://stackoverflow.com/questions/15259720/how-can-i-make-the-microsoft-c-compiler-treat-unknown-flags-as-errors-rather-t
+    def has_argument(self, arg, env):
+        warning_text = b'9002'
+        code = 'int i;\n'
+        (fd, srcname) = tempfile.mkstemp(suffix='.'+self.default_suffix)
+        os.close(fd)
+        ofile = open(srcname, 'w')
+        ofile.write(code)
+        ofile.close()
+        # Read c_args/cpp_args/etc from the cross-info file (if needed)
+        extra_args = self.get_cross_extra_flags(env, compile=True, link=False)
+        extra_args += self.get_compile_only_args()
+        commands = self.exelist + [arg] + extra_args + [srcname]
+        mlog.debug('Running VS compile:')
+        mlog.debug('Command line: ', ' '.join(commands))
+        mlog.debug('Code:\n', code)
+        p = subprocess.Popen(commands, cwd=os.path.split(srcname)[0], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        (stde, stdo) = p.communicate()
+        if p.returncode != 0:
+            raise MesonException('Compiling test app failed.')
+        return not(warning_text in stde or warning_text in stdo)
+
 class VisualStudioCPPCompiler(VisualStudioCCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrap):
         VisualStudioCCompiler.__init__(self, exelist, version, is_cross, exe_wrap)
@@ -1491,10 +1657,16 @@ class GnuCCompiler(CCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch'],
                           '3' : ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch']}
-        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
+                             'b_colorout']
         if self.gcc_type != GCC_OSX:
             self.base_options.append('b_lundef')
 
+    def get_colorout_args(self, colortype):
+        if mesonlib.version_compare(self.version, '>=4.9.0'):
+            return gnu_color_args[colortype][:]
+        return []
+
     def get_pic_args(self):
         if self.gcc_type == GCC_MINGW:
             return [] # On Window gcc defaults to fpic being always on.
@@ -1674,10 +1846,16 @@ class GnuCPPCompiler(CPPCompiler):
         self.warn_args = {'1': ['-Wall', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '2': ['-Wall', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor'],
                           '3': ['-Wall', '-Wpedantic', '-Wextra', '-Winvalid-pch', '-Wnon-virtual-dtor']}
-        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage']
+        self.base_options = ['b_pch', 'b_lto', 'b_pgo', 'b_sanitize', 'b_coverage',
+                             'b_colorout']
         if self.gcc_type != GCC_OSX:
             self.base_options.append('b_lundef')
 
+    def get_colorout_args(self, colortype):
+        if mesonlib.version_compare(self.version, '>=4.9.0'):
+            return gnu_color_args[colortype][:]
+        return []
+
     def get_always_args(self):
         return ['-pipe']
 
@@ -1695,7 +1873,8 @@ class GnuCPPCompiler(CPPCompiler):
 
     def get_options(self):
         opts = {'cpp_std' : coredata.UserComboOption('cpp_std', 'C++ language standard to use',
-                                                     ['none', 'c++03', 'c++11', 'c++14', 'c++1z'],
+                                                     ['none', 'c++03', 'c++11', 'c++14', 'c++1z',
+                                                      'gnu++03', 'gnu++11', 'gnu++14', 'gnu++1z'],
                                                      'none'),
                 'cpp_debugstl': coredata.UserBooleanOption('cpp_debugstl',
                                                            'STL debug mode',
@@ -1796,7 +1975,7 @@ class FortranCompiler(Compiler):
     def needs_static_linker(self):
         return True
 
-    def sanity_check(self, work_dir):
+    def sanity_check(self, work_dir, environment):
         source_name = os.path.join(work_dir, 'sanitycheckf.f90')
         binary_name = os.path.join(work_dir, 'sanitycheckf')
         ofile = open(source_name, 'w')
@@ -1805,7 +1984,8 @@ class FortranCompiler(Compiler):
 end program prog
 ''')
         ofile.close()
-        pc = subprocess.Popen(self.exelist + [source_name, '-o', binary_name])
+        extra_flags = self.get_cross_extra_flags(environment, compile=True, link=True)
+        pc = subprocess.Popen(self.exelist + extra_flags + [source_name, '-o', binary_name])
         pc.wait()
         if pc.returncode != 0:
             raise EnvironmentException('Compiler %s can not compile programs.' % self.name_string())
@@ -1896,6 +2076,14 @@ class GnuFortranCompiler(FortranCompiler):
     def get_always_args(self):
         return ['-pipe']
 
+    def gen_import_library_args(self, implibname):
+        """
+        The name of the outputted import library
+
+        Used only on Windows
+        """
+        return ['-Wl,--out-implib=' + implibname]
+
 class G95FortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper=None)
@@ -1907,6 +2095,14 @@ class G95FortranCompiler(FortranCompiler):
     def get_always_args(self):
         return ['-pipe']
 
+    def gen_import_library_args(self, implibname):
+        """
+        The name of the outputted import library
+
+        Used only on Windows
+        """
+        return ['-Wl,--out-implib=' + implibname]
+
 class SunFortranCompiler(FortranCompiler):
     def __init__(self, exelist, version, is_cross, exe_wrapper=None):
         super().__init__(exelist, version, is_cross, exe_wrapper=None)
@@ -2060,10 +2256,10 @@ class VisualStudioLinker():
         return []
 
     def unix_link_flags_to_native(self, args):
-        return args
+        return args[:]
 
     def unix_compile_flags_to_native(self, args):
-        return args
+        return args[:]
 
 class ArLinker():
 
@@ -2109,7 +2305,7 @@ class ArLinker():
         return []
 
     def unix_link_flags_to_native(self, args):
-        return args
+        return args[:]
 
     def unix_compile_flags_to_native(self, args):
-        return args
+        return args[:]
