@@ -15,7 +15,7 @@
 import pickle, os, uuid
 from .mesonlib import MesonException, default_libdir, default_libexecdir, default_prefix
 
-version = '0.32.0'
+version = '0.33.0'
 backendlist = ['ninja', 'vs2010', 'vs2015', 'xcode']
 
 class UserOption:
@@ -100,10 +100,10 @@ class UserStringArrayOption(UserOption):
                 raise MesonException('Valuestring does not define an array: ' + newvalue)
             newvalue = eval(newvalue, {}, {}) # Yes, it is unsafe.
         if not isinstance(newvalue, list):
-            raise MesonException('String array value is not an array.')
+            raise MesonException('"{0}" should be a string array, but it is not'.format(str(newvalue)))
         for i in newvalue:
             if not isinstance(i, str):
-                raise MesonException('String array element not a string.')
+                raise MesonException('String array element "{0}" is not a string.'.format(str(newvalue)))
         self.value = newvalue
 
 # This class contains all data that must persist over multiple
@@ -133,6 +133,8 @@ class CoreData():
         self.cross_compilers = {}
         self.deps = {}
         self.modules = {}
+        # Only to print a warning if it changes between Meson invocations.
+        self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
 
     def init_builtins(self, options):
         self.builtins = {}
@@ -198,7 +200,7 @@ def get_builtin_option_default(optname):
         raise RuntimeError('Tried to get the default value for an unknown builtin option \'%s\'.' % optname)
 
 builtin_options = {
-        'buildtype'         : [ UserComboOption, 'Build type to use.', [ 'plain', 'debug', 'debugoptimized', 'release' ], 'debug' ],
+        'buildtype'         : [ UserComboOption, 'Build type to use.', [ 'plain', 'debug', 'debugoptimized', 'release', 'minsize' ], 'debug' ],
         'strip'             : [ UserBooleanOption, 'Strip targets on install.', False ],
         'unity'             : [ UserBooleanOption, 'Unity build.', False ],
         'prefix'            : [ UserStringOption, 'Installation prefix.', default_prefix() ],
@@ -228,8 +230,9 @@ forbidden_target_names = {'clean': None,
                           'PHONY': None,
                           'all': None,
                           'test': None,
+                          'test:': None,
                           'test-valgrind': None,
-                          'test-': None,
+                          'test-valgrind:': None,
                           'benchmark': None,
                           'install': None,
                           'build.ninja': None,
