@@ -88,6 +88,8 @@ class PkgConfigDependency(Dependency):
         self.static = kwargs.get('static', False)
         if not isinstance(self.static, bool):
             raise DependencyException('Static keyword must be boolean')
+        self.cargs = []
+        self.libs = []
         if 'native' in kwargs and environment.is_cross_build():
             want_cross = not kwargs['native']
         else:
@@ -100,8 +102,6 @@ class PkgConfigDependency(Dependency):
         if not PkgConfigDependency.pkgconfig_found:
             if self.required:
                 raise DependencyException('Pkg-config not found.')
-            self.cargs = []
-            self.libs = []
             return
         if environment.is_cross_build() and want_cross:
             if "pkgconfig" not in environment.cross_info.config["binaries"]:
@@ -122,58 +122,66 @@ class PkgConfigDependency(Dependency):
             if self.required:
                 raise DependencyException('%s dependency %s not found.' % (self.type_string, name))
             self.modversion = 'none'
-            self.cargs = []
-            self.libs = []
+            return
+        self.modversion = out.decode().strip()
+        found_msg = ['%s dependency' % self.type_string, mlog.bold(name), 'found:']
+        self.version_requirement = kwargs.get('version', None)
+        if self.version_requirement is None:
+            self.is_found = True
         else:
-            self.modversion = out.decode().strip()
-            mlog.log('%s dependency' % self.type_string, mlog.bold(name), 'found:',
-                     mlog.green('YES'), self.modversion)
-            self.version_requirement = kwargs.get('version', None)
-            if self.version_requirement is None:
-                self.is_found = True
-            else:
-                if not isinstance(self.version_requirement, str):
-                    raise DependencyException('Version argument must be string.')
-                self.is_found = mesonlib.version_compare(self.modversion, self.version_requirement)
-                if not self.is_found and self.required:
+            if not isinstance(self.version_requirement, str):
+                raise DependencyException('Version argument must be string.')
+            self.is_found = mesonlib.version_compare(self.modversion, self.version_requirement)
+            if not self.is_found:
+                found_msg += [mlog.red('NO'), 'found {!r}'.format(self.modversion),
+                              'but need {!r}'.format(self.version_requirement)]
+                mlog.log(*found_msg)
+                if self.required:
                     raise DependencyException(
                         'Invalid version of a dependency, needed %s %s found %s.' %
                         (name, self.version_requirement, self.modversion))
-            if not self.is_found:
                 return
-            p = subprocess.Popen([pkgbin, '--cflags', name], stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
-            out = p.communicate()[0]
-            if p.returncode != 0:
-                raise DependencyException('Could not generate cargs for %s:\n\n%s' % \
-                                          (name, out.decode(errors='ignore')))
-            self.cargs = out.decode().split()
+        found_msg += [mlog.green('YES'), self.modversion]
+        mlog.log(*found_msg)
+        # Fetch cargs to be used while using this dependency
+        self._set_cargs()
+        # Fetch the libraries and library paths needed for using this
+        self._set_libs()
 
-            libcmd = [pkgbin, '--libs']
-            if self.static:
-                libcmd.append('--static')
-            p = subprocess.Popen(libcmd + [name], stdout=subprocess.PIPE,
-                                 stderr=subprocess.PIPE)
-            out = p.communicate()[0]
-            if p.returncode != 0:
-                raise DependencyException('Could not generate libs for %s:\n\n%s' % \
-                                          (name, out.decode(errors='ignore')))
-            self.libs = []
-            for lib in out.decode().split():
-                if lib.endswith(".la"):
-                    shared_libname = self.extract_libtool_shlib(lib)
-                    shared_lib = os.path.join(os.path.dirname(lib), shared_libname)
-                    if not os.path.exists(shared_lib):
-                        shared_lib = os.path.join(os.path.dirname(lib), ".libs", shared_libname)
+    def _set_cargs(self):
+        p = subprocess.Popen([self.pkgbin, '--cflags', self.name],
+                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        out = p.communicate()[0]
+        if p.returncode != 0:
+            raise DependencyException('Could not generate cargs for %s:\n\n%s' % \
+                                      (name, out.decode(errors='ignore')))
+        self.cargs = out.decode().split()
 
-                    if not os.path.exists(shared_lib):
-                        raise DependencyException('Got a libtools specific "%s" dependencies'
-                                                  'but we could not compute the actual shared'
-                                                  'library path' % lib)
-                    lib = shared_lib
-                    self.is_libtool = True
+    def _set_libs(self):
+        libcmd = [self.pkgbin, '--libs']
+        if self.static:
+            libcmd.append('--static')
+        p = subprocess.Popen(libcmd + [self.name],
+                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        out = p.communicate()[0]
+        if p.returncode != 0:
+            raise DependencyException('Could not generate libs for %s:\n\n%s' % \
+                                      (name, out.decode(errors='ignore')))
+        self.libs = []
+        for lib in out.decode().split():
+            if lib.endswith(".la"):
+                shared_libname = self.extract_libtool_shlib(lib)
+                shared_lib = os.path.join(os.path.dirname(lib), shared_libname)
+                if not os.path.exists(shared_lib):
+                    shared_lib = os.path.join(os.path.dirname(lib), ".libs", shared_libname)
 
-                self.libs.append(lib)
+                if not os.path.exists(shared_lib):
+                    raise DependencyException('Got a libtools specific "%s" dependencies'
+                                              'but we could not compute the actual shared'
+                                              'library path' % lib)
+                lib = shared_lib
+                self.is_libtool = True
+            self.libs.append(lib)
 
     def get_variable(self, variable_name):
         p = subprocess.Popen([self.pkgbin, '--variable=%s' % variable_name, self.name],
@@ -258,12 +266,13 @@ class WxDependency(Dependency):
 
     def __init__(self, environment, kwargs):
         Dependency.__init__(self)
+        self.is_found = False
         if WxDependency.wx_found is None:
             self.check_wxconfig()
-
         if not WxDependency.wx_found:
-            raise DependencyException('Wx-config not found.')
-        self.is_found = False
+            mlog.log("Neither wx-config-3.0 nor wx-config found; can't detect dependency")
+            return
+
         p = subprocess.Popen([self.wxc, '--version'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
@@ -344,37 +353,13 @@ class WxDependency(Dependency):
 class ExternalProgram():
     def __init__(self, name, fullpath=None, silent=False, search_dir=None):
         self.name = name
-        self.fullpath = None
         if fullpath is not None:
             if not isinstance(fullpath, list):
                 self.fullpath = [fullpath]
             else:
                 self.fullpath = fullpath
         else:
-            self.fullpath = [shutil.which(name)]
-            if self.fullpath[0] is None and search_dir is not None:
-                trial = os.path.join(search_dir, name)
-                suffix = os.path.splitext(trial)[-1].lower()[1:]
-                if mesonlib.is_windows() and (suffix == 'exe' or suffix == 'com'\
-                                          or suffix == 'bat'):
-                    self.fullpath = [trial]
-                elif not mesonlib.is_windows() and os.access(trial, os.X_OK):
-                    self.fullpath = [trial]
-                else:
-                    # Now getting desperate. Maybe it is a script file that is a) not chmodded
-                    # executable or b) we are on windows so they can't be directly executed.
-                    try:
-                        first_line = open(trial).readline().strip()
-                        if first_line.startswith('#!'):
-                            commands = first_line[2:].split('#')[0].strip().split()
-                            if mesonlib.is_windows():
-                                # Windows does not have /usr/bin.
-                                commands[0] = commands[0].split('/')[-1]
-                                if commands[0] == 'env':
-                                    commands = commands[1:]
-                            self.fullpath = commands + [trial]
-                    except Exception:
-                        pass
+            self.fullpath = self._search(name, search_dir)
         if not silent:
             if self.found():
                 mlog.log('Program', mlog.bold(name), 'found:', mlog.green('YES'),
@@ -382,6 +367,67 @@ class ExternalProgram():
             else:
                 mlog.log('Program', mlog.bold(name), 'found:', mlog.red('NO'))
 
+    @staticmethod
+    def _shebang_to_cmd(script):
+        """
+        Windows does not understand shebangs, so we check if the file has a
+        shebang and manually parse it to figure out the interpreter to use
+        """
+        try:
+            first_line = open(script).readline().strip()
+            if first_line.startswith('#!'):
+                commands = first_line[2:].split('#')[0].strip().split()
+                if mesonlib.is_windows():
+                    # Windows does not have /usr/bin.
+                    commands[0] = commands[0].split('/')[-1]
+                    if commands[0] == 'env':
+                        commands = commands[1:]
+                return commands + [script]
+        except Exception:
+            pass
+        return False
+
+    @staticmethod
+    def _is_executable(path):
+        suffix = os.path.splitext(path)[-1].lower()[1:]
+        if mesonlib.is_windows():
+            if suffix == 'exe' or suffix == 'com' or suffix == 'bat':
+                return True
+        elif os.access(path, os.X_OK):
+            return True
+        return False
+
+    def _search_dir(self, name, search_dir):
+        if search_dir is None:
+            return False
+        trial = os.path.join(search_dir, name)
+        if not os.path.exists(trial):
+            return False
+        if self._is_executable(trial):
+            return [trial]
+        # Now getting desperate. Maybe it is a script file that is a) not chmodded
+        # executable or b) we are on windows so they can't be directly executed.
+        return self._shebang_to_cmd(trial)
+
+    def _search(self, name, search_dir):
+        commands = self._search_dir(name, search_dir)
+        if commands:
+            return commands
+        # Do a standard search in PATH
+        fullpath = shutil.which(name)
+        if fullpath or not mesonlib.is_windows():
+            # On UNIX-like platforms, the standard PATH search is enough
+            return [fullpath]
+        # On Windows, interpreted scripts must have an extension otherwise they
+        # cannot be found by a standard PATH search. So we do a custom search
+        # where we manually search for a script with a shebang in PATH.
+        search_dirs = os.environ.get('PATH', '').split(';')
+        for search_dir in search_dirs:
+            commands = self._search_dir(name, search_dir)
+            if commands:
+                return commands
+        return [None]
+
     def found(self):
         return self.fullpath[0] is not None
 
@@ -1016,6 +1062,7 @@ class SDL2Dependency(Dependency):
             self.linkargs = stdo.decode().strip().split()
             self.is_found = True
             mlog.log('Dependency', mlog.bold('sdl2'), 'found:', mlog.green('YES'), '(%s)' % sdlconf)
+            self.version = '2' # FIXME
             return
         try:
             pcdep = PkgConfigDependency('sdl2', kwargs)
@@ -1023,6 +1070,7 @@ class SDL2Dependency(Dependency):
                 self.is_found = True
                 self.cargs = pcdep.get_compile_args()
                 self.linkargs = pcdep.get_link_args()
+                self.version = pcdep.get_version()
                 return
         except Exception:
             pass
@@ -1032,6 +1080,7 @@ class SDL2Dependency(Dependency):
                 self.is_found = True
                 self.cargs = fwdep.get_compile_args()
                 self.linkargs = fwdep.get_link_args()
+                self.version = '2' # FIXME
                 return
         mlog.log('Dependency', mlog.bold('sdl2'), 'found:', mlog.red('NO'))
 
@@ -1044,6 +1093,9 @@ class SDL2Dependency(Dependency):
     def found(self):
         return self.is_found
 
+    def get_version(self):
+        return self.version
+
 class ExtraFrameworkDependency(Dependency):
     def __init__(self, name, required, path=None):
         Dependency.__init__(self)
@@ -1100,11 +1152,13 @@ class Python3Dependency(Dependency):
         super().__init__()
         self.name = 'python3'
         self.is_found = False
+        self.version = "3.something_maybe"
         try:
             pkgdep = PkgConfigDependency('python3', environment, kwargs)
             if pkgdep.found():
                 self.cargs = pkgdep.cargs
                 self.libs = pkgdep.libs
+                self.version = pkgdep.get_version()
                 self.is_found = True
                 return
         except Exception:
@@ -1122,6 +1176,7 @@ class Python3Dependency(Dependency):
                 self.libs = ['-L{}/libs'.format(basedir),
                              '-lpython{}'.format(vernum)]
                 self.is_found = True
+                self.version = sysconfig.get_config_var('py_version_short')
             elif mesonlib.is_osx():
                 # In OSX the Python 3 framework does not have a version
                 # number in its name.
@@ -1141,6 +1196,9 @@ class Python3Dependency(Dependency):
     def get_link_args(self):
         return self.libs
 
+    def get_version(self):
+        return self.version
+
 def get_dep_identifier(name, kwargs):
     elements = [name]
     modlist = kwargs.get('modules', [])
