@@ -45,11 +45,13 @@ def find_valgrind():
 def detect_ninja():
     for n in ['ninja', 'ninja-build']:
         try:
-            p = subprocess.Popen([n, '--version'], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
+            p = subprocess.Popen([n, '--version'], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)
         except FileNotFoundError:
             continue
-        p.communicate()
-        if p.returncode == 0:
+        version = p.communicate()[0].decode(errors='ignore')
+        # Perhaps we should add a way for the caller to know the failure mode
+        # (not found or too old)
+        if p.returncode == 0 and mesonlib.version_compare(version, ">=1.6"):
             return n
 
 def detect_cpu_family():
@@ -80,11 +82,37 @@ def detect_system():
     return platform.system().lower()
 
 
+def for_windows(is_cross, env):
+    """
+    Host machine is windows?
+
+    Note: 'host' is the machine on which compiled binaries will run
+    """
+    if not is_cross:
+        return mesonlib.is_windows()
+    elif env.cross_info.has_host():
+        return env.cross_info.config['host_machine']['system'] == 'windows'
+    return False
+
+def for_darwin(is_cross, env):
+    """
+    Host machine is Darwin (iOS/OS X)?
+
+    Note: 'host' is the machine on which compiled binaries will run
+    """
+    if not is_cross:
+        return mesonlib.is_osx()
+    elif env.cross_info.has_host():
+        return env.cross_info.config['host_machine']['system'] == 'darwin'
+    return False
+
+
 class Environment():
     private_dir = 'meson-private'
     log_dir = 'meson-logs'
     coredata_file = os.path.join(private_dir, 'coredata.dat')
     version_regex = '\d+(\.\d+)+(-[a-zA-Z0-9]+)?'
+
     def __init__(self, source_dir, build_dir, main_script_file, options, original_cmd_line_args):
         assert(os.path.isabs(main_script_file))
         assert(not os.path.islink(main_script_file))
@@ -100,7 +128,11 @@ class Environment():
             self.coredata = coredata.load(cdf)
             self.first_invocation = False
         except FileNotFoundError:
+            # WARNING: Don't use any values from coredata in __init__. It gets
+            # re-initialized with project options by the interpreter during
+            # build file parsing.
             self.coredata = coredata.CoreData(options)
+            self.coredata.meson_script_file = self.meson_script_file
             self.first_invocation = True
         if self.coredata.cross_file:
             self.cross_info = CrossBuildInfo(self.coredata.cross_file)
@@ -122,40 +154,27 @@ class Environment():
         self.default_static_linker = 'ar'
         self.vs_static_linker = 'lib'
 
+        # Various prefixes and suffixes for import libraries, shared libraries,
+        # static libraries, and executables.
+        # Versioning is added to these names in the backends as-needed.
         cross = self.is_cross_build()
         if (not cross and mesonlib.is_windows()) \
         or (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'windows'):
             self.exe_suffix = 'exe'
-            if self.detect_c_compiler(cross).get_id() == 'msvc':
-                self.import_lib_suffix = 'lib'
-            else:
-                # MinGW-GCC doesn't generate and can't link with a .lib
-                # It uses the DLL file as the import library
-                self.import_lib_suffix = 'dll'
-            self.shared_lib_suffix = 'dll'
-            self.shared_lib_prefix = ''
-            self.static_lib_suffix = 'lib'
-            self.static_lib_prefix = ''
             self.object_suffix = 'obj'
+            self.win_libdir_layout = True
         else:
             self.exe_suffix = ''
-            if (not cross and mesonlib.is_osx()) or \
-            (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'darwin'):
-                self.shared_lib_suffix = 'dylib'
-            else:
-                self.shared_lib_suffix = 'so'
-            self.shared_lib_prefix = 'lib'
-            self.static_lib_suffix = 'a'
-            self.static_lib_prefix = 'lib'
             self.object_suffix = 'o'
-            self.import_lib_suffix = self.shared_lib_suffix
+            self.win_libdir_layout = False
 
     def is_cross_build(self):
         return self.cross_info is not None
 
-    def dump_coredata(self):
+    def dump_coredata(self, mtime):
         cdf = os.path.join(self.get_build_dir(), Environment.coredata_file)
         coredata.save(self.coredata, cdf)
+        os.utime(cdf, times=(mtime, mtime))
 
     def get_script_dir(self):
         import mesonbuild.scripts
@@ -246,7 +265,7 @@ class Environment():
                 version = 'unknown version'
             if 'apple' in out and 'Free Software Foundation' in out:
                 return GnuCCompiler(ccache + [compiler], version, GCC_OSX, is_cross, exe_wrap)
-            if (out.startswith('cc') or 'gcc' in out) and \
+            if (out.startswith('cc') or 'gcc' in out.lower()) and \
                 'Free Software Foundation' in out:
                 lowerout = out.lower()
                 if 'mingw' in lowerout or 'msys' in lowerout or 'mingw' in compiler.lower():
@@ -641,22 +660,19 @@ class Environment():
     def get_exe_suffix(self):
         return self.exe_suffix
 
-    # On Windows (MSVC) the library has suffix dll
-    # but you link against a file that has suffix lib.
-    def get_import_lib_suffix(self):
-        return self.import_lib_suffix
+    def get_import_lib_dir(self):
+        "Install dir for the import library (library used for linking)"
+        return self.get_libdir()
 
-    def get_shared_lib_prefix(self):
-        return self.shared_lib_prefix
+    def get_shared_lib_dir(self):
+        "Install dir for the shared library"
+        if self.win_libdir_layout:
+            return self.get_bindir()
+        return self.get_libdir()
 
-    def get_shared_lib_suffix(self):
-        return self.shared_lib_suffix
-
-    def get_static_lib_prefix(self):
-        return self.static_lib_prefix
-
-    def get_static_lib_suffix(self):
-        return self.static_lib_suffix
+    def get_static_lib_dir(self):
+        "Install dir for the static library"
+        return self.get_libdir()
 
     def get_object_suffix(self):
         return self.object_suffix
@@ -682,17 +698,6 @@ class Environment():
     def get_datadir(self):
         return self.coredata.get_builtin_option('datadir')
 
-    def find_library(self, libname, dirs):
-        if dirs is None:
-            dirs = mesonlib.get_library_dirs()
-        suffixes = [self.get_shared_lib_suffix(), self.get_static_lib_suffix()]
-        prefix = self.get_shared_lib_prefix()
-        for d in dirs:
-            for suffix in suffixes:
-                trial = os.path.join(d, prefix + libname + '.' + suffix)
-                if os.path.isfile(trial):
-                    return trial
-
 
 def get_args_from_envvars(lang):
     if lang == 'c':
@@ -721,14 +726,12 @@ def get_args_from_envvars(lang):
 
 class CrossBuildInfo():
     def __init__(self, filename):
-        self.config = {}
+        self.config = {'properties': {}}
         self.parse_datafile(filename)
         if 'target_machine' in self.config:
             return
         if not 'host_machine' in self.config:
             raise mesonlib.MesonException('Cross info file must have either host or a target machine.')
-        if not 'properties' in self.config:
-            raise mesonlib.MesonException('Cross file is missing "properties".')
         if not 'binaries' in self.config:
             raise mesonlib.MesonException('Cross file is missing "binaries".')
 
@@ -771,6 +774,9 @@ class CrossBuildInfo():
     def get_stdlib(self, language):
         return self.config['properties'][language + '_stdlib']
 
+    def get_properties(self):
+        return self.config['properties']
+
     # Wehn compiling a cross compiler we use the native compiler for everything.
     # But not when cross compiling a cross compiler.
     def need_cross_compiler(self):
