@@ -15,7 +15,7 @@
 # limitations under the License.
 
 import sys, stat, traceback, pickle, argparse
-import datetime
+import time, datetime
 import os.path
 from . import environment, interpreter, mesonlib
 from . import build
@@ -117,6 +117,13 @@ itself as required.'''
                 raise RuntimeError('Something went terribly wrong. Please file a bug.')
         return (src_dir, build_dir)
 
+    def check_pkgconfig_envvar(self, env):
+        curvar = os.environ.get('PKG_CONFIG_PATH', '')
+        if curvar != env.coredata.pkgconf_envvar:
+            mlog.log(mlog.red("WARNING:"), 'PKG_CONFIG_PATH has changed between invocations from "%s" to "%s".' %
+                     (env.coredata.pkgconf_envvar, curvar))
+            env.coredata.pkgconf_envvar = curvar
+
     def generate(self):
         env = environment.Environment(self.source_dir, self.build_dir, self.meson_script_file, self.options, self.original_cmd_line_args)
         mlog.initialize(env.get_log_dir())
@@ -124,6 +131,7 @@ itself as required.'''
         mlog.debug('Python binary:', sys.executable)
         mlog.debug('Python system:', platform.system())
         mlog.log(mlog.bold('The Meson build system'))
+        self.check_pkgconfig_envvar(env)
         mlog.log('Version:', coredata.version)
         mlog.log('Source dir:', mlog.bold(self.source_dir))
         mlog.log('Build dir:', mlog.bold(self.build_dir))
@@ -156,11 +164,21 @@ itself as required.'''
         mlog.log('Build machine cpu family:', mlog.bold(intr.builtin['build_machine'].cpu_family_method([], {})))
         mlog.log('Build machine cpu:', mlog.bold(intr.builtin['build_machine'].cpu_method([], {})))
         intr.run()
-        env.dump_coredata()
+        coredata_mtime = time.time()
         g.generate(intr)
         g.run_postconf_scripts()
         dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
         pickle.dump(b, open(dumpfile, 'wb'))
+        # Write this last since we use the existence of this file to check if
+        # we generated the build file successfully, so we don't want an error
+        # that pops up during generation, post-conf scripts, etc to cause us to
+        # incorrectly signal a successful meson run which will cause an error
+        # about an already-configured build directory when the user tries again.
+        #
+        # However, we set the mtime to an earlier value to ensure that doing an
+        # mtime comparison between the coredata dump and other build files
+        # shows the build files to be newer, not older.
+        env.dump_coredata(coredata_mtime)
 
 def run_script_command(args):
     cmdname = args[0]
